from matchescu.matching.evaluation.data.benchmark import (
    MagellanBenchmarkData,
    MagellanTraits,
)
from matchescu.matching.evaluation.data.extraction._record_extraction import (
    CsvRecordExtraction,
)


__all__ = ["MagellanBenchmarkData", "MagellanTraits", "CsvRecordExtraction"]



import itertools
import logging
from contextlib import AbstractContextManager
from pathlib import Path

import numpy as np
import torch
from sklearn import metrics
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from matchescu.matching.matchers.ml.ditto._ditto_module import DittoModel

from matchescu.matching.matchers.ml.ditto.training._datasets import DittoDataset


class TrainingEvaluator(AbstractContextManager):

    def __init__(
        self,
        task_name: str,
        xv_data: DataLoader[DittoDataset],
        test_data: DataLoader[DittoDataset],
        tb_log_dir: Path,
        logger: logging.Logger | None = None,
    ) -> None:
        self._task = task_name
        self._xv_data = xv_data
        self._test_data = test_data
        self._best_test_f1 = 0.0
        self._best_xv_f1 = 0.0
        self._tb_log_dir = tb_log_dir.absolute()
        self._summary_writer = SummaryWriter(log_dir=str(self._tb_log_dir))
        self._log = (logger or logging.getLogger(self.__class__.__name__)).getChild(
            self._task
        )

    @property
    def summary_writer(self):
        return self._summary_writer

    @staticmethod
    def __best_threshold(
        probabilities: np.ndarray, labels: np.ndarray
    ) -> tuple[float, float]:
        thresholds = np.arange(0.0, 1.0, 0.05)
        predictions = (probabilities[:, None] > thresholds).astype(int)
        f1_scores = np.fromiter(
            itertools.starmap(
                metrics.f1_score, zip(itertools.repeat(labels), predictions.T)
            ),
            dtype=np.float32,
        )
        best_idx = np.argmax(f1_scores)
        return float(f1_scores[best_idx]), float(thresholds[best_idx])

    @staticmethod
    @torch.no_grad()
    def _evaluate_model(
        model: DittoModel, data_loader: DataLoader, threshold: float | None = None
    ):
        model.eval()

        batch_results = map(lambda b: (torch.sigmoid(model(b[0])), b[1]), data_loader)
        all_probs, all_y = zip(*batch_results)
        all_probs = torch.cat(all_probs).detach().cpu().numpy()
        all_y = torch.cat(all_y).detach().cpu().numpy()

        if threshold is not None:
            pred = (all_probs > threshold).astype(int)
            f1 = metrics.f1_score(all_y, pred)
            return f1, threshold
        return TrainingEvaluator.__best_threshold(all_probs, all_y)

    def __call__(
        self, model: DittoModel, train_metrics: dict, epoch: int
    ) -> tuple[bool, float]:
        self._log.info("evaluating on cross-validation set")
        xv_f1, best_xv_threshold = self._evaluate_model(model, self._xv_data)
        self._log.info(
            "X validation F1=%.4f, best threshold=%.2f",
            xv_f1,
            best_xv_threshold,
        )
        test_f1, _ = self._evaluate_model(
            model, self._test_data, threshold=best_xv_threshold
        )
        self._log.info("test F1=%.4f", test_f1)
        train_metrics.update({"dev F1": xv_f1, "test F1": test_f1})
        self._summary_writer.add_scalars(self._task, train_metrics, epoch)
        found_new_best = False

        if xv_f1 > self._best_xv_f1:
            self._log.info("found new best F1. saving checkpoint")
            self._best_xv_f1 = xv_f1
            self._best_test_f1 = test_f1

            self._log.info(
                "xv_f1=%.4f, best_xv_f1=%.4f, test_f1=%.4f, best_test_f1=%.4f",
                xv_f1,
                self._best_xv_f1,
                test_f1,
                self._best_test_f1,
            )
            found_new_best = True

        return found_new_best, best_xv_threshold

    def __enter__(self) -> "TrainingEvaluator":
        if self._summary_writer is not None:
            return self
        self._summary_writer = SummaryWriter(log_dir=str(self._tb_log_dir))
        return self

    def __exit__(self, exc_type, exc_value, traceback, /):
        if self._summary_writer is None:
            return

        self._summary_writer.flush()
        self._summary_writer.close()
        self._summary_writer = None



import sys
import time
import warnings
import logging
from contextlib import contextmanager
from datetime import timedelta
from pathlib import Path
from typing import Iterable

import click
import humanize
from transformers import PreTrainedTokenizerBase, AutoTokenizer

from matchescu.data import Record
from matchescu.extraction import Traits
from matchescu.matching.evaluation.data import MagellanBenchmarkData, MagellanTraits
from matchescu.matching.matchers.ml.deepmatcher import DeepMatcherModule
from matchescu.matching.matchers.ml.deepmatcher.training._config import (
    TrainingConfig,
    DEFAULT_MODEL_DIR,
    DEFAULT_DATA_DIR,
)
from matchescu.matching.matchers.ml.deepmatcher.training._dataset import (
    DeepMatcherDataset,
)
from matchescu.matching.matchers.ml.deepmatcher.training._evaluator import (
    TrainingEvaluator,
)
from matchescu.matching.matchers.ml.deepmatcher.training._trainer import (
    DeepMatcherTrainer,
)
from matchescu.typing import (
    EntityReferenceIdentifier as RefId,
)

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger("deepmatcher-training")


@contextmanager
def timer(start_message: str):
    log.info(start_message)
    time_start = time.time()
    yield
    time_end = time.time()
    duration = humanize.naturaldelta(timedelta(seconds=(time_end - time_start)))
    log.info("%s time elapsed: %s", start_message, duration)


@timer(start_message="load dataset")
def load_magellan_dataset(
    ds_path: Path, left_traits: Traits, right_traits: Traits | None = None
) -> MagellanBenchmarkData:
    ds = MagellanBenchmarkData(ds_path)
    ds.load_left(left_traits)
    ds.load_right(right_traits or left_traits)
    ds.load_splits()
    return ds


@timer(start_message="serialize+tokenize")
def get_datasets(
    magellan_ds: MagellanBenchmarkData,
    tokenizer: PreTrainedTokenizerBase,
) -> tuple[DeepMatcherDataset, ...]:
    return tuple(
        DeepMatcherDataset(magellan_ds.id_table, split, tokenizer=tokenizer)
        for split in [
            magellan_ds.train_split,
            magellan_ds.valid_split,
            magellan_ds.test_split,
        ]
    )


@timer(start_message="train deepmatcher")
def train_on_benchmark_data(
    model_save_dir: Path,
    model_name: str,
    benchmark_data: MagellanBenchmarkData,
    tokenizer: PreTrainedTokenizerBase,
    batch_size: int = 32,
    epochs: int = 20,
    learning_rate: float = 1e-5,
):
    vocab_size = len(tokenizer.get_vocab())
    train_ds, xv_ds, test_ds = get_datasets(benchmark_data, tokenizer)
    train, xv, test = (
        train_ds.get_data_loader(batch_size, shuffle=True),
        xv_ds.get_data_loader(batch_size * 16),
        test_ds.get_data_loader(batch_size * 16),
    )
    model = DeepMatcherModule(vocab_size, train_ds.attr_dims, train_ds.attr_count)
    dataset_logger = log.getChild(benchmark_data.name)
    trainer = DeepMatcherTrainer(
        model_name,
        model_save_dir,
        epochs=epochs,
        learning_rate=learning_rate,
        logger=dataset_logger,
    )
    tb_log_dir = model_save_dir / model_name / "tensorboard"
    with TrainingEvaluator(
        model_name, xv, test, tb_log_dir, dataset_logger
    ) as evaluator:
        trainer.run_training(model, train, evaluator, True)


def table_a_id(rows: Iterable[Record]) -> RefId:
    return RefId(next(iter(rows))["id"], "tableA")


def table_b_id(rows: Iterable[Record]) -> RefId:
    return RefId(next(iter(rows))["id"], "tableB")


@click.command
@click.option(
    "-M",
    "--model-dir",
    "root_model_dir",
    required=True,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=DEFAULT_MODEL_DIR,
)
@click.option(
    "-D",
    "--dataset-dir",
    "root_data_dir",
    required=True,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=DEFAULT_DATA_DIR,
)
@click.option(
    "-f",
    "--config-file",
    "config_path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    default=DEFAULT_MODEL_DIR / "config.json",
)
def run_training(
    root_model_dir: Path,
    root_data_dir: Path,
    config_path: Path,
) -> None:
    root_model_dir = Path(root_model_dir)
    root_data_dir = Path(root_data_dir)
    benchmark_dataset_traits = MagellanTraits()
    config = TrainingConfig.load_json(config_path)
    model_name = "dm-hybrid-symmetric-attention"
    with warnings.catch_warnings(action="ignore"):
        for dataset_name in config.dataset_names:
            ds_path = root_data_dir / "magellan" / dataset_name
            ds_model_dir = root_model_dir / dataset_name
            tokenizer = AutoTokenizer.from_pretrained("google-bert/bert-base-uncased")
            train_params = config.get(dataset=dataset_name)
            ds_traits = benchmark_dataset_traits[dataset_name]
            benchmark_data = load_magellan_dataset(ds_path, ds_traits)
            train_on_benchmark_data(
                ds_model_dir,
                model_name,
                benchmark_data,
                tokenizer,
                batch_size=train_params.batch_size,
                epochs=train_params.epochs,
                learning_rate=train_params.learning_rate,
            )


if __name__ == "__main__":
    run_training()



import json
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import Optional, cast

DEFAULT_MODEL_DIR = Path.cwd() / "models"
DEFAULT_DATA_DIR = Path.cwd() / "data"


@dataclass
class TrainingParams:
    learning_rate: float = field(default=3e-5)
    epochs: int = field(default=10)
    batch_size: int = field(default=32)


@dataclass
class TrainingConfig(TrainingParams):
    dataset_names: list[str] = field(default_factory=list)
    dataset_configs: dict[str, TrainingParams] = field(default_factory=dict)

    @classmethod
    def load_json(cls, f: str | PathLike) -> "TrainingConfig":
        with open(f, "r") as fp:
            config = json.load(fp)
        result = TrainingConfig()
        result.dataset_names = list(map(str, config.get("datasets", [])))
        result.learning_rate = float(config.get("learningRate", 3e-5))
        result.batch_size = int(config.get("batchSize", 32))
        result.epochs = int(config.get("epochs", 15))

        for ds_name, params in config.get("options", {}).items():
            result.dataset_configs[ds_name] = TrainingParams(
                learning_rate=float(params.get("learningRate", result.learning_rate)),
                batch_size=int(params.get("batchSize", result.batch_size)),
                epochs=int(params.get("epochs", result.epochs)),
            )
        return result

    @staticmethod
    def __attr_val[T](params: TrainingParams, attr_name: str) -> T:
        return cast(T, params.__getattribute__(attr_name))

    def _get_attr[T](self, attr_name: str, dataset: str, default_val: T) -> T:
        cfg = self.dataset_configs.get(dataset, self)
        val = self.__attr_val(cfg, attr_name)
        return val if val is not None and val != default_val else default_val

    def get(self, dataset: Optional[str] = None) -> TrainingParams:
        return self.dataset_configs.get(dataset, self)



import logging
from contextlib import AbstractContextManager
from pathlib import Path

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from sklearn.metrics import f1_score

from matchescu.matching.matchers.ml.ditto.training._datasets import DittoDataset


class TrainingEvaluator(AbstractContextManager):
    def __init__(
        self,
        task_name: str,
        xv_data: DataLoader[DittoDataset],
        test_data: DataLoader[DittoDataset],
        tb_log_dir: Path,
        logger: logging.Logger | None = None,
    ) -> None:
        self._task = task_name
        self._xv_data = xv_data
        self._test_data = test_data
        self._best_test_f1 = 0.0
        self._best_xv_f1 = 0.0
        self._tb_log_dir = tb_log_dir.absolute()
        self._summary_writer = SummaryWriter(log_dir=str(self._tb_log_dir))
        self._log = (logger or logging.getLogger(self.__class__.__name__)).getChild(
            self._task
        )

    @property
    def summary_writer(self):
        return self._summary_writer

    @staticmethod
    @torch.no_grad()
    def _evaluate_model(model: nn.Module, data_loader: DataLoader):
        model.eval()
        y_pred = []
        y_true = []

        for batch in data_loader:
            left_attrs = batch["left_attrs"]
            right_attrs = batch["right_attrs"]
            labels = batch["label"].long()

            logits = model(left_attrs, right_attrs)
            y_pred.extend(logits.argmax(dim=1).detach().cpu().numpy())
            y_true.extend(labels.detach().cpu().numpy())

        return f1_score(y_true, y_pred)

    def __call__(self, model: nn.Module, epoch: int) -> tuple[bool, float]:
        self._log.info("evaluating on cross-validation set")
        xv_f1 = self._evaluate_model(model, self._xv_data)
        self._log.info("dev F1=%.4f", xv_f1)
        test_f1 = self._evaluate_model(model, self._test_data)
        self._log.info("test F1=%.4f", test_f1)
        self._summary_writer.add_scalars(
            self._task, {"dev F1": xv_f1, "test F1": test_f1}, epoch
        )

        found_new_best = False
        if xv_f1 > self._best_xv_f1:
            self._log.info("found new best F1. saving checkpoint")
            self._best_xv_f1 = xv_f1
            self._best_test_f1 = test_f1

            self._log.info(
                "xv_f1=%.4f, best_xv_f1=%.4f, test_f1=%.4f, best_test_f1=%.4f",
                xv_f1,
                self._best_xv_f1,
                test_f1,
                self._best_test_f1,
            )
            found_new_best = True

        return found_new_best, xv_f1

    def __enter__(self) -> "TrainingEvaluator":
        if self._summary_writer is not None:
            return self
        self._summary_writer = SummaryWriter(log_dir=str(self._tb_log_dir))
        return self

    def __exit__(self, exc_type, exc_value, traceback, /):
        if self._summary_writer is None:
            return

        self._summary_writer.flush()
        self._summary_writer.close()
        self._summary_writer = None



import json
from dataclasses import dataclass, field, asdict
from os import PathLike
from pathlib import Path
from typing import Type, Optional, cast

DEFAULT_MODEL_DIR = Path.cwd() / "models"
DEFAULT_DATA_DIR = Path.cwd() / "data"


@dataclass
class ModelTrainingParams:
    learning_rate: float = field(default=3e-5)
    frozen_layer_count: int = field(default=0)
    epochs: int = field(default=10)
    batch_size: int = field(default=32)


@dataclass
class DatasetTrainingParams(ModelTrainingParams):
    model_configs: dict[str, ModelTrainingParams] = field(default_factory=dict)


@dataclass
class TrainingConfig(DatasetTrainingParams):
    model_names: list[str] = field(default_factory=list)
    dataset_names: list[str] = field(default_factory=list)
    dataset_configs: dict[str, DatasetTrainingParams] = field(default_factory=dict)

    @classmethod
    def __read_model_training_params[T: ModelTrainingParams](
        cls, config: dict, parent: T | None, clazz: Type[T]
    ) -> T:
        lr = 3e-5 if parent is None else parent.learning_rate
        fl = 0 if parent is None else parent.frozen_layer_count
        ep = 10 if parent is None else parent.epochs
        bs = 32 if parent is None else parent.batch_size
        return clazz(
            learning_rate=(float(config.get("learningRate", lr))),
            frozen_layer_count=(int(config.get("frozenLayers", fl))),
            epochs=(int(config.get("epochs", ep))),
            batch_size=(int(config.get("batchSize", bs))),
        )

    @classmethod
    def __read_dataset_training_params[T: DatasetTrainingParams](
        cls, config: dict, parent: T | None, clazz: Type[T]
    ) -> T:
        result = cls.__read_model_training_params(config, parent, clazz)
        if "modelConfig" in config:
            result.model_configs = {
                key: cls.__read_model_training_params(
                    config_node, result, ModelTrainingParams
                )
                for key, config_node in config["modelConfig"].items()
            }
        return result

    @classmethod
    def load_json(cls, f: str | PathLike) -> "TrainingConfig":
        with open(f, "r") as fp:
            config = json.load(fp)

        result = cls.__read_dataset_training_params(config, None, TrainingConfig)
        result.dataset_names = list(map(str, config.get("datasets", [])))
        result.model_names = list(map(str, config.get("models", [])))
        if "datasetConfig" in config:
            result.dataset_configs = {
                key: cls.__read_dataset_training_params(
                    config_node, result, DatasetTrainingParams
                )
                for key, config_node in config["datasetConfig"].items()
            }
        return result

    @staticmethod
    def __attr_val[T](params: ModelTrainingParams, attr_name: str) -> T:
        return cast(T, params.__getattribute__(attr_name))

    def _get_attr[T](
        self, attr_name: str, model: str, dataset: str, default_val: T
    ) -> T:
        cfg = self.dataset_configs.get(dataset, self)
        val = self.__attr_val(cfg, attr_name)
        return val if val is not None and val != default_val else default_val

    def get(
        self,
        model: Optional[str] = None,
        dataset: Optional[str] = None,
    ) -> ModelTrainingParams:
        cfg = self.dataset_configs.get(dataset, self)
        if model is None:
            return cfg

        if model in cfg.model_configs:
            return cfg.model_configs[model]

        if model in self.model_configs:
            base_cfg = asdict(self)
            local_cfg = asdict(cfg)
            remote_cfg = asdict(self.model_configs[model])
            out = {}
            for k, v in local_cfg.items():
                if k == "model_configs":
                    continue
                base_v = base_cfg.get(k, v)
                remote_v = remote_cfg.get(k, v)
                if remote_v != v and remote_v != base_v:
                    out[k] = remote_v
                else:
                    out[k] = v
            cfg = ModelTrainingParams(**out)
        return cfg



from abc import ABC
from pathlib import Path
from typing import Iterable


class BenchmarkData(ABC):
    @staticmethod
    def _check_files(paths: list[str | Path]) -> Iterable[Path]:
        for path in map(Path, paths):
            if not path.is_file():
                raise FileNotFoundError(path)
            yield path.absolute()



from logging import Logger, getLogger
from os import PathLike
from pathlib import Path
from typing import Any, cast

import torch
from torch import nn
from torch.nn.modules.loss import _Loss
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader
from transformers import get_linear_schedule_with_warmup

from matchescu.matching.matchers.ml.deepmatcher.training._dataset import (
    DeepMatcherDataset,
)
from matchescu.matching.matchers.ml.deepmatcher.training._evaluator import (
    TrainingEvaluator,
)


class DeepMatcherTrainer:
    def __init__(
        self,
        task_name: str,
        model_dir: str | PathLike | None = None,
        loss_fn: _Loss | None = None,
        **kwargs: Any,
    ) -> None:
        self._task = task_name
        self._model_dir = Path(model_dir) if model_dir else Path(__file__).parent
        self._log = cast(
            Logger, kwargs.get("logger", getLogger(self.__class__.__name__))
        ).getChild(self._task)
        self._epochs = int(kwargs.get("epochs", 20))
        self._learning_rate = float(kwargs.get("learning_rate", 3e-5))
        self._loss = loss_fn or nn.CrossEntropyLoss()

    def __get_device(self):
        device = torch.device("cpu")
        if torch.cuda.is_available():
            device = torch.device(f"cuda:{torch.cuda.current_device()}")
        elif torch.backends.mps.is_available():
            device = torch.device("mps:0")
        else:
            if torch.backends.mps.is_built():
                self._log.info("MPS built, but not available.")
            else:
                self._log.info("Not Mac, nor CUDA.")
        return device

    def _train_one_epoch(
        self,
        epoch: int,
        device: torch.device,
        model: torch.nn.Module,
        train_iter: DataLoader,
        optimizer: Optimizer,
        scheduler: LRScheduler,
    ):
        total_loss = 0.0
        batch_no = 0

        try:
            loss_fn = self._loss.to(device)
            model.to(device)
            model.train(True)
            batch_loss = 0.0
            for i, batch in enumerate(train_iter):
                inputs = {k: v.to(device) for k, v in batch.items() if k != "label"}
                labels = batch["label"].to(device)
                optimizer.zero_grad()

                prediction = model(**inputs)
                loss = loss_fn(prediction, labels)

                loss.backward()
                optimizer.step()
                scheduler.step()

                step_loss = loss.item()
                total_loss += step_loss
                batch_loss += step_loss
                batch_no = i + 1
                if batch_no % 10 == 0:
                    batch_loss = batch_loss / 10
                    fmt = f"batch {batch_no}: avg loss over last 10 batches=%.4f"
                    self._log.info(fmt, batch_loss)
                del loss
        finally:
            model.train(False)

        avg_loss = total_loss / batch_no if batch_no > 0 else 0
        self._log.info("epoch %d: avg loss=%.4f", epoch, avg_loss)
        return {"Average Loss": avg_loss}

    def run_training(
        self,
        model: nn.Module,
        training_data: DataLoader[DeepMatcherDataset],
        evaluator: TrainingEvaluator,
        save_model: bool = False,
    ):
        device = self.__get_device()
        optimizer = torch.optim.AdamW(model.parameters(), lr=self._learning_rate)
        ds = cast(DeepMatcherDataset, training_data.dataset)
        batch_count = int(len(ds) // training_data.batch_size)
        num_steps = batch_count * self._epochs
        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=0, num_training_steps=num_steps
        )

        for epoch in range(1, self._epochs + 1):
            self._log.info("epoch %d - train start", epoch)
            try:
                self._train_one_epoch(
                    epoch,
                    device,
                    model,
                    training_data,
                    optimizer,
                    scheduler,
                )
            finally:
                self._log.info("epoch %d - train end", epoch)

            if evaluator is None or not save_model:
                continue

            is_new_best, dev_f1 = evaluator(model, epoch)
            if is_new_best:
                self._save_checkpoint(epoch, model, optimizer, scheduler)

    def _save_checkpoint(
        self,
        epoch: int,
        model: nn.Module,
        optimizer: Optimizer,
        scheduler: LRScheduler,
    ):
        task_model_dir = self._model_dir / self._task
        task_model_dir.mkdir(parents=True, exist_ok=True)
        ckpt_path = task_model_dir / "model.pt"
        ckpt = {
            "model": model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "scheduler": scheduler.state_dict(),
            "epoch": epoch,
        }
        torch.save(ckpt, ckpt_path)



import time
import warnings
from contextlib import contextmanager
from datetime import timedelta
from functools import partial
from pathlib import Path

import click
import humanize
from torch.utils.data import DataLoader
from transformers import AutoTokenizer, PreTrainedTokenizerFast, DebertaV2TokenizerFast

from matchescu.extraction import Traits
from matchescu.matching.evaluation.data import MagellanBenchmarkData, MagellanTraits
from matchescu.matching.matchers.ml.ditto._ditto_module import DittoModel
from matchescu.matching.matchers.ml.ditto.training._config import (
    TrainingConfig,
    ModelTrainingParams,
    DEFAULT_MODEL_DIR,
    DEFAULT_DATA_DIR,
)
from matchescu.matching.matchers.ml.ditto.training._datasets import DittoDataset
from matchescu.matching.matchers.ml.ditto.training._trainer import DittoTrainer
from matchescu.matching.matchers.ml.ditto.training._evaluator import TrainingEvaluator
from matchescu.matching.matchers.ml.ditto.training._logging import log


_MODEL_TOKENIZERS = {
    "microsoft/deberta-v3-base": DebertaV2TokenizerFast.from_pretrained,
}


@contextmanager
def timer(start_message: str):
    log.info(start_message)
    time_start = time.time()
    yield
    time_end = time.time()
    duration = humanize.naturaldelta(timedelta(seconds=(time_end - time_start)))
    log.info("%s time elapsed: %s", start_message, duration)


@timer(start_message="load dataset")
def load_magellan_dataset(
    ds_path: Path,
    left_traits: Traits,
    right_traits: Traits | None = None,
) -> MagellanBenchmarkData:
    ds = MagellanBenchmarkData(ds_path)
    ds.load_left(left_traits)
    ds.load_right(right_traits or left_traits)
    ds.load_splits()
    return ds


@timer(start_message="serialize+tokenize")
def get_magellan_data_loaders(
    magellan_ds: MagellanBenchmarkData,
    tokenizer: PreTrainedTokenizerFast,
    batch_size: int = 32,
) -> tuple[DataLoader, DataLoader, DataLoader]:
    train_ds, xv_ds, test_ds = [
        DittoDataset(magellan_ds.id_table, split, tokenizer)
        for split in [
            magellan_ds.train_split,
            magellan_ds.valid_split,
            magellan_ds.test_split,
        ]
    ]
    return (
        train_ds.get_data_loader(batch_size, shuffle=True),
        train_ds.get_data_loader(batch_size * 16),
        train_ds.get_data_loader(batch_size * 16),
    )


@timer(start_message="train ditto")
def train_on_benchmark_data(
    model_save_dir: Path,
    model_name: str,
    benchmark_data: MagellanBenchmarkData,
    tokenizer: PreTrainedTokenizerFast,
    train_params: ModelTrainingParams,
):
    train, xv, test = get_magellan_data_loaders(
        benchmark_data, tokenizer, train_params.batch_size
    )
    ditto = DittoModel(model_name)
    dataset_logger = log.getChild(benchmark_data.name)
    trainer = DittoTrainer(
        model_name,
        model_save_dir,
        learning_rate=train_params.learning_rate,
        epochs=train_params.epochs,
        logger=dataset_logger,
    )
    tb_log_dir = model_save_dir / model_name / "tensorboard"
    with TrainingEvaluator(
        model_name, xv, test, tb_log_dir, dataset_logger
    ) as evaluator:
        trainer.run_training(ditto, train, evaluator, True)


@click.command
@click.option(
    "-M",
    "--model-dir",
    "root_model_dir",
    required=True,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=DEFAULT_MODEL_DIR,
)
@click.option(
    "-D",
    "--dataset-dir",
    "root_data_dir",
    required=True,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=DEFAULT_DATA_DIR,
)
@click.option(
    "-f",
    "--config-file",
    "config_path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    default=DEFAULT_MODEL_DIR / "config.json",
)
def run_training(
    root_model_dir: Path,
    root_data_dir: Path,
    config_path: Path,
) -> None:
    root_model_dir = Path(root_model_dir)
    root_data_dir = Path(root_data_dir)
    benchmark_dataset_traits = MagellanTraits()
    config = TrainingConfig.load_json(config_path)
    with warnings.catch_warnings(action="ignore"):
        for dataset_name in config.dataset_names:
            ds_path = root_data_dir / "magellan" / dataset_name
            ds_model_dir = root_model_dir / dataset_name
            for model_name in config.model_names:
                tokenizer = _new_fast_tokenizer(model_name)
                train_params = config.get(model=model_name, dataset=dataset_name)
                ds_traits = benchmark_dataset_traits[dataset_name]
                benchmark_data = load_magellan_dataset(ds_path, ds_traits)

                train_on_benchmark_data(
                    ds_model_dir,
                    model_name,
                    benchmark_data,
                    tokenizer,
                    train_params,
                )


def _new_fast_tokenizer(model_name: str) -> PreTrainedTokenizerFast:
    tokenizer_factory = _MODEL_TOKENIZERS.get(
        model_name, partial(AutoTokenizer.from_pretrained, use_fast=True)
    )
    return tokenizer_factory(model_name)


if __name__ == "__main__":
    run_training()
